/* 
ahb_dma.c programmed by Ivan Wang
AHB dma program for audio 2004/8/13 06:00pm

AHB dma function usage:
        // First, you must allocate structure and call init
            ahb_dma_data_t  *priv;
            priv=ahb_dma_alloc();
            //prepare variables for initialize
            priv->base=CPE_AHBDMA_VA_BASE;          //base addr
            priv->llp_master=AHBDMA_MASTER_0;       //LLP master channel number
            priv->src_data_master=AHBDMA_MASTER_0;  //src master channel number
            priv->dest_data_master=AHBDMA_MASTER_0; //dst master channel number
            priv->llp_count=LLP_COUNT;              //LLP number for link list
            priv->channel=PMU_SSP_DMA_CHANNEL;      //DMA channel number
            priv->hw_handshake=1;                   //if hardware handshake?
            ahb_dma_init(priv);
            
        // Second, add dma entry
            //prepare parameter for add dma entry
            ahb_dma_parm_t parm;
            parm.src=SRC_ADDR;                      //given source phy addr
            parm.dest=DEST_ADDR;                    //given dest phy addr
            parm.sw=AHBDMA_WIDTH_32BIT;             //source width 32 bit
            parm.dw=AHBDMA_WIDTH_32BIT;             //dest width 32 bit
            parm.sctl=AHBDMA_CTL_FIX;               //source fix control
            parm.dctl=AHBDMA_CTL_FIX;               //dest fix control
            parm.size=DMA_SIZE;                     //transfer count
            parm.irq=AHBDMA_NO_TRIGGER_IRQ;         //if irq trigger?
            ahb_dma_add(priv,&parm);                //call add function

            ....you may use ahb_dma_add (LLP_COUNT+1) times
            
        // Third, start dma
            ahb_dma_start(priv);

        //clear interrupt
            ahb_dma_clear_int(priv);
                                
        // Last, Free structure
            ahb_dma_free(priv);
            
        // useful function for you to maintain
            ahb_dma_reset(priv); //reset dma action and stop
            
            addr=ahbdma_get_dest_addr(priv); get DMA destinition address
            
            status=ahbdma_get_status(priv); get dma status,Bit0:finished IRQ   Bit1:Error IRQ

Note:
****Example to Playing I2S 3 LLP Count dma and its dma size is less 0x800
    (first time)
    +-----------+        +----------+      +----------+       +----------+ 
    |DMA reg    | ---->  | Link 1   | ---> | Link 2   | --->  | Link 3   | ---> NULL
    |NULL data  |        | NULL data|      | NULL data|       | Real data| 
    |No Trig IRQ|        | Trig IRQ |      | Trig IRQ |       | Trig IRQ | 
    +-----------+        +----------+      +----------+       +----------+ 

    (second time,Link 1 Finished and irq triggered)
         +----------+       +----------+       +----------+ 
         | Link 2   | --->  | Link 3   | --->  | Link 1   |  ---> NULL
         | NULL data|       | Real data|       | Real data| 
         | Trig IRQ |       | Trig IRQ |       | Trig IRQ | 
         +----------+       +----------+       +----------+ 
         
    (third time,Link 2 Finished and irq triggered)
         +----------+       +----------+       +----------+ 
         | Link 3   | --->  | Link 1   | --->  | Link 2   |  ---> NULL
         | Real data|       | Real data|       | Real data| 
         | Trig IRQ |       | Trig IRQ |       | Trig IRQ | 
         +----------+       +----------+       +----------+ 

****Example to Playing I2S 3 LLP Count dma and its dma size is 0x1000,you must use

    (first time)
    +-----------+        +----------+      +----------+       +----------+      +----------+     
    |DMA reg    | ---->  | Link 1   | ---> | Link 2   | --->  | Link 3-1 | ---> | Link 3-2 | --->NULL
    |NULL data  |        | NULL data|      | NULL data|       |0x800 size|      |0x800 size|
    |No Trig IRQ|        | Trig IRQ |      | Trig IRQ |       | Real data|      | Real data|    
    +-----------+        +----------+      +----------+       | Trig IRQ |      | Trig IRQ |     
                                                              +----------+      +----------+

****Example to Recording I2S 3 LLP Count dma and its dma size is 0x1000,you must use

    (first time)   
    +-----------+      +----------+      +----------+     +----------+      +----------+     +----------+      +----------+          
    |DMA reg    | ---> | Link 1-1 | ---> | Link 1-2 | --->| Link 2-1 | ---> | Link 2-2 | --->| Link 3-1 | ---> | Link 2-2 | ---> NULL 
    |NULL data  |      |0x800 size|      |0x800 size|     |0x800 size|      |0x800 size|     |0x800 size|      |0x800 size|          
    |No Trig IRQ|      | Real data|      | Real data|     | Real data|      | Real data|     | Real data|      | Real data|          
    +-----------+      | NoTrigIRQ|      | Trig IRQ |     | NoTrigIRQ|      | Trig IRQ |     | NoTrigIRQ|      | Trig IRQ |          
                       +----------+      +----------+     +----------+      +----------+     +----------+      +----------+


*/

#include <linux/module.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <asm/types.h>
#include <asm/io.h>
#include <asm/arch/ahb_dma.h>


ahb_dma_data_t *ahb_dma_alloc(void)
{
    ahb_dma_data_t *priv;
    priv=(ahb_dma_data_t *)kmalloc(sizeof(ahb_dma_data_t), GFP_KERNEL);
    return priv;
}

void ahb_dma_free(ahb_dma_data_t *priv)
{
    int size;
    size=sizeof(ahb_lld_t)*(priv->llp_count);
    if(priv->ahb_dma_lld)
    {
        //printk("free 0x%x with 0x%x size 0x%x\n",priv->ahb_dma_lld,priv->ahb_dma_lld_phys,size);
        consistent_free((void *)priv->ahb_dma_lld,size,(dma_addr_t)priv->ahb_dma_lld_phys);
    }       
    ahb_dma_reset(priv);
    kfree(priv);
}

/* 
You must given:
    base:   base address of ahb dma
    llp_master: LLP master channel number
    src_data_master: source data channel number
    dest_data_master: dest data channel number
    llp_count:  LLP count
    channel:    ahb dma channel
    hw      is it hardware handshake?
*/
int ahb_dma_init(ahb_dma_data_t *priv)
{   
    int             i,size;
    ahb_lld_t       *lld=0;
    ahb_lld_t       *lld_phys=0;
    
    if(priv==0)
    {
        printk("NULL priv data!\n");
        return 0;
    }
    
    if(priv->llp_count)
    {
        size=sizeof(ahb_lld_t)*(priv->llp_count);
        //printk("allocate size=0x%x sizeof(ahb_lld_t)=0x%x\n",size,sizeof(ahb_lld_t));
        priv->ahb_dma_lld=(ahb_lld_t *)consistent_alloc(GFP_DMA|GFP_KERNEL,
            size,(dma_addr_t *)&priv->ahb_dma_lld_phys);
        //printk("priv->ahb_dma_lld=0x%x phys=0x%x size=0x%x\n",(int)priv->ahb_dma_lld,(int)priv->ahb_dma_lld_phys,size);
        if(priv->ahb_dma_lld==0)
        {
            printk("ahb_dma_lld allocate memory fail!\n");
            return 0;
        }
    }
    
    priv->channel_base=(priv->base)+0x100+((priv->channel)*0x20);
    priv->llp_last_idx=0;
    priv->llp_free_idx=0;
    priv->ahb_last_lld=0;
    
    lld=priv->ahb_dma_lld;
    lld_phys=priv->ahb_dma_lld_phys;
    /* initialize value */
    for(i=0;i<priv->llp_count;i++)
    {
//printk("lld[%d]=0x%x\n",i,&lld[i]);
        lld[i].source=0;  //source
        lld[i].dest=0;     //dest,16bit
        lld[i].control=0;
        
        if(i==priv->llp_count-1)
            lld[i].llp=0;
        else
            lld[i].llp=(unsigned int)(&lld_phys[i+1])|(priv->llp_master); //use second channel 
        //printk("ahb_dma_lld[%d].llp\n",i,ahb_dma_lld[i].llp);
    }

    *(unsigned int *)(priv->channel_base+AHBDMA_CHANNEL_LLP_0x10)=0;
    *(unsigned int *)(priv->channel_base+AHBDMA_CHANNEL_CFG_0x4)=0x0;    //enabled all interrupt
    *(unsigned int *)(priv->base+AHBDMA_ENABLE_0x24)=0x1;  //enable DMA  controller   
    *(unsigned int *)(priv->channel_base+AHBDMA_CHANNEL_CSR_0x0)=0;
    return 1;
}

/*
LLP count =3 (example)
   0 => 1 => 2 => 3
        ^         |
        +---------+
You must prepared:
    src: source address
    dest:dest address
    sw: source width (0/1/2=>8/16/32)
    dw: dest width (0/1/2=>8/16/32)
    sctl: source control (0/1/2/3=>inc/dec/fix/x)
    dctl: dest coontrol (0/1/2/3=>inc/dec/fix/x)
    size: dma count
    irq: (0/1)==>(disable/enable)
*/
void ahb_dma_add(ahb_dma_data_t *priv,ahb_dma_parm_t *parm)
{
    unsigned int        val;
    volatile ahb_lld_t  *lld;
    ahb_lld_t           *lld_phys;
    //printk("add priv=0x%x\n",priv);
    lld=priv->ahb_dma_lld;
    lld_phys=priv->ahb_dma_lld_phys;

//printk("add lld=0x%x priv->llp_count=%d priv->llp_free_idx=%d parm->size=0x%x\n",lld,priv->llp_count,priv->llp_free_idx,parm->size);
//printk("dest=0x%x sz=0x%x\n",parm->dest,parm->size);
    if(priv->llp_free_idx==0) //first to call ahb_dma_add
    {
        __asm__("nop");__asm__("nop");__asm__("nop");
        *(volatile unsigned int *)(priv->channel_base+AHBDMA_CHANNEL_TXSZ_0x14)=parm->size;
        *(volatile unsigned int *)(priv->channel_base+AHBDMA_CHANNEL_SRC_0x8)=parm->src;
        *(volatile unsigned int *)(priv->channel_base+AHBDMA_CHANNEL_DST_0xC)=parm->dest;

        if (priv->llp_count == 0)
           *(volatile unsigned int *)(priv->channel_base+AHBDMA_CHANNEL_LLP_0x10)= 0;  //john modified, for no LLP
		else
           *(volatile unsigned int *)(priv->channel_base+AHBDMA_CHANNEL_LLP_0x10)=
               (volatile unsigned int)(&lld_phys[0])|priv->llp_master;
        
        val=(((1-(parm->irq))<<31)&0x80000000)|
            ((parm->sw<<11)&0x00003800)|((parm->dw<<8)&0x00000700)|
            ((priv->hw_handshake<<7)&0x00000080)|
            ((parm->sctl<<5)&0x00000060)|((parm->dctl<<3)&0x00000018)|
            ((priv->src_data_master<<2)&0x4)|
            ((priv->dest_data_master<<1)&0x2);
        *(volatile unsigned int *)(priv->channel_base+AHBDMA_CHANNEL_CSR_0x0)=val;
    }
    else
    {
//printk("parm=0x%x\n",&parm);
        val=(((1-(parm->irq))<<28)&0x10000000)|
            ((parm->sw<<25)&0x0e000000)|((parm->dw<<22)&0x01c00000)|
            ((parm->sctl<<20)&0x00300000)|((parm->dctl<<18)&0x000c0000)|
            ((priv->src_data_master<<17)&0x00020000)|
            ((priv->dest_data_master<<16)&0x00010000)|(parm->size);
            
        lld[priv->llp_free_idx-1].source=parm->src;
        lld[priv->llp_free_idx-1].dest=parm->dest;
        lld[priv->llp_free_idx-1].control=val;
        lld[priv->llp_free_idx-1].llp=0;
        if(priv->ahb_last_lld)
            priv->ahb_last_lld->llp=(unsigned int)(&lld_phys[priv->llp_free_idx-1])|priv->llp_master;
        priv->ahb_last_lld=(ahb_lld_t *)&lld[priv->llp_free_idx-1];
    }
    
    if (priv->llp_count)   //john add check
	{
       if(priv->llp_free_idx==priv->llp_count)
          priv->llp_free_idx=1;
       else
          priv->llp_free_idx++;
	}
}

inline void ahb_dma_start(ahb_dma_data_t *priv)
{
    *(unsigned int *)(priv->base+AHBDMA_ENABLE_0x24)=0x1;  //enable DMA  controller
    *(volatile unsigned int *)(priv->channel_base+AHBDMA_CHANNEL_CSR_0x0)|=0x1;
}

void ahb_dma_reset(ahb_dma_data_t *priv)
{
    priv->llp_last_idx=0;
    priv->llp_free_idx=0;
    priv->ahb_last_lld=0;

    //printk("ahb_dma_stop\n");
    *(unsigned int *)(priv->base+AHBDMA_ENABLE_0x24)=0; //disable DMA controller
    *(unsigned int *)(priv->channel_base+AHBDMA_CHANNEL_CSR_0x0)=0;   //disable DMA channel 0         
    *(unsigned int *)(priv->channel_base+AHBDMA_CHANNEL_LLP_0x10)=0;
    *(unsigned int *)(priv->channel_base+AHBDMA_CHANNEL_TXSZ_0x14)=0;  //no transfer size (use LLP)
    *(unsigned int *)(priv->channel_base+AHBDMA_CHANNEL_CFG_0x4)=0;   //disable all interrupt
}

inline void ahbdma_clear_int(ahb_dma_data_t *priv)
{
    *(volatile unsigned int *)(priv->base+AHBDMA_ISR_0x8)=1<<(priv->channel);
}

//john add, get ext DMA address
// so we can get DMA lenth
inline unsigned int ahbdma_get_dest_addr(ahb_dma_data_t *priv)
{
	return *(volatile unsigned int *)(priv->channel_base+AHBDMA_CHANNEL_DST_0xC);
}

// John add, to get interrupt status
// Bit 0 indicate interrupt created, 
// bit 1 indicate interrupt error
unsigned int ahbdma_get_status(ahb_dma_data_t *priv)
{
	u32 status=0;
    u32 temp;

    temp = *(volatile u32 *)(priv->base+AHBDMA_INT_ERR_0x0C);
	if ( temp & (1<<(priv->channel)) ) 
	   status |= INT_ERROR;   //error

    temp = *(volatile u32 *)(priv->base+AHBDMA_INT_TC_0x04);
	if ( temp & (1<<(priv->channel)) )
	   status |= INT_TRIGGER;   //complete

    return status;
}


inline int ahbdma_get_irq(unsigned int base)
{
    unsigned int     status;

    status=*(volatile unsigned int *)(base+AHBDMA_INT_TC_0x04);
    if(status&0x1)
        return 0;
    if(status&0x2)
        return 1;
    if(status&0x4)
        return 2;
    if(status&0x8)
        return 3;
    return -1;
}


EXPORT_SYMBOL(ahb_dma_alloc);
EXPORT_SYMBOL(ahb_dma_free);
EXPORT_SYMBOL(ahb_dma_init);
EXPORT_SYMBOL(ahb_dma_add);
EXPORT_SYMBOL(ahb_dma_start);
EXPORT_SYMBOL(ahb_dma_reset);
EXPORT_SYMBOL(ahbdma_clear_int);
EXPORT_SYMBOL(ahbdma_get_dest_addr);
EXPORT_SYMBOL(ahbdma_get_status);
EXPORT_SYMBOL(ahbdma_get_irq);